#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#ifndef _MSC_VER
#include "config.h"
#endif
#include "WebSrv.h"

#ifdef DEBUG
#define debug printf
#else
#define debug
#endif

// BASE64 decoding routine
char *WebSrv_unbase64(const char *str);

static char *memstr(char *a, int s, char *b)
{
	int l;
	char *p;
	
	l=strlen(b);
	for(p=a;p<a+s-l;p++)
	{
		if(!memcmp(p,b,l))
			return(p);
	}
	return(NULL);
}
#ifdef _MSC_VER
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#else
#ifndef HAVE_STRNCASECMP
static int strncasecmp(char *a, char*b, int n)
{
	char *c,*d,*p;
	int o;
	
	c=malloc(n+1);
	memcpy(c,a,n);
	c[n]='\0';
	for(p=c;*p;p++)
		if(*p>='a'&&*p<='a')
			*p-='a'-'A';
	d=strdup(b);
	for(p=d;*p;p++)
		if(*p>='a'&&*p<='a')
			*p-='a'-'A';
	o=strcmp(c,d);
	free(c);
	free(d);
	return(o);
}
#endif
#endif

WebSrv *WebSrv_Create()
{
	WebSrv *ws;
	
	ws=(WebSrv*)malloc(sizeof(WebSrv));
	memset(ws,0,sizeof(WebSrv));
	return(ws);
}

TCPsocket WebSrv_Open(WebSrv *websrv, Uint16 port)
{
	IPaddress ip;
	SDLNet_ResolveHost(&ip, NULL, port);
	websrv->sock=SDLNet_TCP_Open(&ip);
	if(websrv->sock)
	{
		websrv->set=SDLNet_AllocSocketSet(1);
		SDLNet_TCP_AddSocket(websrv->set, websrv->sock);
	}
	return(websrv->sock);
}

int WebSrv_Poll(WebSrv *websrv, Uint32 timeout)
{
	return(SDLNet_CheckSockets(websrv->set,timeout));
}

char *WebSrv_ClientReadLine(WebSrvClient *client)
{
	static char *buf=NULL;
	int len=0, err;
	char c;
	
	if(buf)
		free(buf);
	buf=NULL;
	while((err=SDLNet_TCP_Recv(client->sock,&c,1))==1)
	{
		if(c=='\r')
			continue;
		buf=(char*)realloc(buf,len+1);
		buf[len]=c;
		if(c=='\n')
		{
			buf[len]='\0';
			return(buf);
		}
		len++;
	}
	if(err<0)
	{
		if(buf)
			free(buf);
		buf=NULL;
	}
	return(buf);
}

void WebSrv_ParseParams(WebSrvURL *url, char *buf)
{
	char *p,*e,*ee;
	e=buf-1;
	do
	{
		url->params=(WebSrvParams*)realloc(url->params,sizeof(WebSrvParams)*(url->num_params+1));
		memset(&url->params[url->num_params],0,sizeof(WebSrvParams));
		p=e+1;
		e=strchr(p,'&');
		if(e)
			*e=0;
		ee=strchr(p,'=');
		if(ee)
			*ee=0;
		url->params[url->num_params].name=WebSrv_dehex(p);
		if(ee)
		{
			url->params[url->num_params].data=WebSrv_dehex(ee+1);
			url->params[url->num_params].size=strlen(url->params[url->num_params].data);
			*ee='=';
		}
		url->num_params++;
		if(e)
			*e='&';
	} while(e);
}

void WebSrv_ParseMultipart(WebSrvClient *c)
{
	char *p,*e,*ee;
	int blen;

	blen=strlen(c->boundary);
	//fwrite(c->post,c->post_length,1,stdout); printf("\n");
	if(strncmp(c->post,"--",2) || strncmp(c->boundary, c->post+2, blen))
	{
		debug("boundary not on first line of Multipart POST!\n");
		return;
	}
	for(p=c->post+2+blen;p;p=e)
	{
		while(*p=='\n' || *p=='\r')
			p++;
		e=p;
		e=memstr(e,c->post_length+(char*)c->post-e,c->boundary);
		if(!e)
		{
			debug("end part boundary not found!\n");
			return;
		}
		e[-4]='\0';

		p=strstr(p,"Content-Disposition: form-data; name=\"");
		if(p)
		{
			p+=38;
			ee=strchr(p,'\"');
			if(ee)
			{
				*ee='\0';
				c->url.params=(WebSrvParams*)realloc(c->url.params,sizeof(WebSrvParams)*(c->url.num_params+1));
				memset(&c->url.params[c->url.num_params],0,sizeof(WebSrvParams));
				c->url.params[c->url.num_params].name=strdup(p);
				c->url.params[c->url.num_params].filename=NULL;
				*ee='\"';
				// filename
				ee=strstr(p,"; filename=\"");
				if(ee)
				{
					p=ee+12;
					ee=strchr(p,'\n');
					if(ee)
						ee=strrchr(p,'\"');
					if(ee)
					{
						*ee='\0';
						c->url.params[c->url.num_params].filename=strdup(p);
						*ee='\"';
					}
				}
				// data
				p=strstr(p,"\r\n\r\n");
				if(p)
				{
					p+=4;
					//debug("+++begin+++\n%s\n---end---\n",p);
					c->url.params[c->url.num_params].size=e-p-4;
					if(c->url.params[c->url.num_params].size)
					{
						c->url.params[c->url.num_params].data=malloc(c->url.params[c->url.num_params].size);
						memcpy(c->url.params[c->url.num_params].data,p,c->url.params[c->url.num_params].size);
					}
				}
				c->url.num_params++;
			}
		}
		
		e[-4]='\r';
		e+=blen;
		if(*e=='-')
			break;
	}
}

int WebSrv_ParseURL(WebSrvURL *url)
{
	char *p,*e,*ee;

	p=url->url;
	if(!strncmp(p,"http:",5))
		p+=5;
	e=strchr(p,'/');
	if(e!=p)
	{
		*e=0;
		ee=strchr(p,':');
		if(ee)
			*ee=0;
		url->host=strdup(p);
		if(ee)
		{
			url->port=strtol(ee+1,NULL,0)&0xFFFF;
			*ee=':';
		}
		*e='/';
	}
	p=e;
	e=strchr(p,'?');
	if(e)
		*e=0;
	ee=strrchr(p,'/');
	if(!ee)
		return(-1); /* this meanes it couldn't find the trailing slash...impossible! */
	*ee=0;
	url->path=strdup(p);
	*ee='/';
	url->file=strdup(ee+1);
	if(e)
	{
		*e='?';
		WebSrv_ParseParams(url,e);
	}
	return(0);
}

WebSrvClient *WebSrv_Accept(WebSrv *websrv)
{
	WebSrvClient *c;
	TCPsocket sock;
	char *buf,*p;
	int ispost=0, gotlen=0;
	
	sock=SDLNet_TCP_Accept(websrv->sock);
	if(!sock)
		return(NULL);
	c=(WebSrvClient*)malloc(sizeof(WebSrvClient));
	memset(c,0,sizeof(WebSrvClient));
	/*memset(&c->url,0,sizeof(WebSrvURL));*/
	websrv->clients=(WebSrvClient**)realloc(websrv->clients,sizeof(WebSrvClient*)*(websrv->num_clients+1));
	websrv->clients[websrv->num_clients]=c;
	websrv->num_clients++;
	c->sock=sock;
	c->url.encoding=WEBSRV_INVALID;
	while((buf=WebSrv_ClientReadLine(c)) && strlen(buf))
	{
		/*debug("line='%s'\n",buf);*/
		if(!strncasecmp(buf,"GET ",4))
		{
			p=strchr(buf+4,' ');
			if(p)
				*p='\0';
			c->url.url=strdup(buf+4);
			/*debug("url=%s\n",c->url.url);*/
			WebSrv_ParseURL(&c->url);
			c->url.encoding=WEBSRV_URL;
		}
		else if(!strncasecmp(buf,"POST ",5))
		{
			ispost=1;
			p=strchr(buf+5,' ');
			if(p)
				*p='\0';
			c->url.url=strdup(buf+5);
			WebSrv_ParseURL(&c->url);
		}
		else if(!strncasecmp(buf,"Content-length: ",16))
		{
			gotlen=1;
			c->post_length=strtol(buf+16,NULL,0);
		}
		else if(!strncasecmp(buf,"Content-type: ",14))
		{
			if(!strncasecmp(buf+14,"application/x-www-form-urlencoded",33))
				c->url.encoding=WEBSRV_URL;
			if(!strncasecmp(buf+14,"multipart/form-data",19))
			{
				char *s,*e;
				if((s=strstr(buf+14+19,"boundary=")))
				{
					c->url.encoding=WEBSRV_MULTIPART;
					e=strpbrk(s+9," \n\r");
					if(e)
						*e='\0';
					c->boundary=strdup(s+9);
				}
			}
		}
		else if(!strncasecmp(buf,"Authorization: Basic ",21))
		{
			c->auth=WebSrv_unbase64(buf+21);
			debug("Basic Authorization: %s\n",c->auth);
		}
		else
			debug("discarded: %s\n",buf);
	}
	if(!buf)
	{
		WebSrv_CloseClient(websrv,c);
		c=NULL;
	}
	if(ispost && c && c->url.encoding!=WEBSRV_INVALID)
	{
		//debug("post: encoding=%d length=%d boundary=\"%s\"\n",c->url.encoding, c->post_length,c->boundary);
		if(!gotlen)
		{
			debug("POST with no Content-Length!\n");
			WebSrv_SendHTTPResponse(c,400);
			WebSrv_CloseClient(websrv,c);
			c=NULL;
		}
		else if(c->post_length)
		{
			c->post=(unsigned char*)malloc(c->post_length+1);
			if(!c->post)
			{
				debug("POST Content-length too big for memory!\n");
				WebSrv_SendHTTPResponse(c,500); /* out of memory */
				WebSrv_CloseClient(websrv,c);
				c=NULL;
			}
			else
			{
				memset(c->post,0,c->post_length+1);
				gotlen=0;
				while(gotlen!=c->post_length)
				{
					int len;
					if((len=SDLNet_TCP_Recv(c->sock,c->post+gotlen,c->post_length-gotlen))<0)
					{
						debug("POST Content-length not received in full!\n");
						c->post_length=gotlen;
					}
					else
						gotlen+=len;
				}
				if(c->url.encoding==WEBSRV_URL)
					WebSrv_ParseParams(&c->url,c->post);
				else if(c->url.encoding==WEBSRV_MULTIPART && c->boundary)
					WebSrv_ParseMultipart(c);
			}
		}
	}
	return(c);
}

char *WebSrv_enhex(const char *str)
{
    const char *p;
    char *s,*out;

    out=(char*)malloc(strlen(str)*3+1);
    for(p=str,s=out;*p;p++,s++)
    {
	if(*p==' ')
		*s='+';
	else if(!strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz*_/-.+",*p))
        {
            sprintf(s,"%%%02X",*p);
            s+=2;
        }
        else
            *s=*p;
    }
    *s='\0';
    return(out);
}

char *WebSrv_dehex(const char *str)
{
    const char *p;
    char *out,*s,hex[3];

    out=s=(char*)malloc(strlen(str)+1);
    hex[2]=0;
    for(p=str;*p;p++,s++)
    {
        if(*p=='%')
        {
            hex[0]=p[1];
            hex[1]=p[2];
            *s=(char)strtol(hex,NULL,16);
			p+=2;
        }
	else if(*p=='+')
		*s=' ';
	else
		*s=*p;
    }
    *s=0;
    out=realloc(out,strlen(out)+1);
    return(out);
}

int WebSrv_ClientSendVString(WebSrvClient *client, char *format, ...)
{
	char str[4096];
	va_list ap;

	va_start(ap, format);
	vsprintf (str, format, ap);
	va_end(ap);
	str[4095]=0;
	return(WebSrv_ClientSendString(client, str));
}

int WebSrv_SendHTTPResponse(WebSrvClient *client, Uint32 code)
{
	switch(code)
	{
		/* 1XX are not in the standard, they are informational only */
		/* 2XX */
		case 200:
			WebSrv_ClientSendString(client,"HTTP/1.0 200 OK\r\n");
			break;
		case 201:
			WebSrv_ClientSendString(client,"HTTP/1.0 201 Created\r\n");
			break;
		case 202:
			WebSrv_ClientSendString(client,"HTTP/1.0 202 Accepted\r\n");
			break;
		case 204:
			WebSrv_ClientSendString(client,"HTTP/1.0 204 No Content\r\n");
			break;
		/* 3XX */
		case 300:
			WebSrv_ClientSendString(client,"HTTP/1.0 300 Multiple Choices\r\n");
			break;
		case 301:
			WebSrv_ClientSendString(client,"HTTP/1.0 301 Moved Permanently\r\n");
			break;
		case 302:
			WebSrv_ClientSendString(client,"HTTP/1.0 302 Moved Temporarily\r\n");
			break;
		case 304:
			WebSrv_ClientSendString(client,"HTTP/1.0 304 Not Modified\r\n");
			break;
		/* 4XX */
		case 400:
			WebSrv_ClientSendString(client,"HTTP/1.0 400 Bad Request\r\n");
			break;
		case 401:
			WebSrv_ClientSendString(client,"HTTP/1.0 401 Unauthorized\r\n");
			break;
		case 403:
			WebSrv_ClientSendString(client,"HTTP/1.0 403 Forbidden\r\n");
			break;
		case 404:
			WebSrv_ClientSendString(client,"HTTP/1.0 404 Not Found\r\n");
			break;
		/* 5XX */
		case 500:
			WebSrv_ClientSendString(client,"HTTP/1.0 500 Internal Server Error\r\n");
			break;
		case 501:
			WebSrv_ClientSendString(client,"HTTP/1.0 501 Not Implemented\r\n");
			break;
		case 503:
			WebSrv_ClientSendString(client,"HTTP/1.0 503 Service Unavailable\r\n");
			break;
		default:
			return(-1);
			break;
	}
	return(0);
}

void WebSrv_FreeURL(WebSrvURL *url)
{
	int i;

	if(url->url)
		free(url->url);
	if(url->host)
		free(url->host);
	if(url->path)
		free(url->path);
	if(url->file)
		free(url->file);
	if(url->params)
	{
		for(i=0;i<url->num_params;i++)
		{
			if(url->params[i].name)
				free(url->params[i].name);
			if(url->params[i].filename)
				free(url->params[i].filename);
			if(url->params[i].data) /* yes, this can be NULL, Virginia! */
				free(url->params[i].data);
		}
		free(url->params);
	}
	memset(url,0,sizeof(WebSrvURL));
}

void WebSrv_CloseClient(WebSrv *websrv, WebSrvClient *client)
{
	int i;

	if(!client)
		return;
	for(i=0; i<websrv->num_clients && client!=websrv->clients[i]; i++);
	if(!i<websrv->num_clients)
		return;
	SDLNet_TCP_Close(client->sock);
	WebSrv_FreeURL(&client->url);
	if(client->post)
		free(client->post);
	if(client->boundary)
		free(client->boundary);
        if(client->auth)
		free(client->auth);
	free(client);
	memmove(&websrv->clients[i], &websrv->clients[i+1], sizeof(WebSrvClient*)*(websrv->num_clients-i-1));
	websrv->num_clients--;
	if(websrv->num_clients-i)
		websrv->clients=(WebSrvClient**)realloc(websrv->clients,sizeof(WebSrvClient*)*(websrv->num_clients-i));
	else
	{
		if(websrv->clients)
			free(websrv->clients);
		websrv->clients=NULL;
	}
}

void WebSrv_Close(WebSrv *websrv)
{
	SDLNet_TCP_Close(websrv->sock);
	websrv->sock=NULL;
	while(websrv->num_clients)
		WebSrv_CloseClient(websrv, websrv->clients[0]);
	websrv->num_clients=0;
	SDLNet_FreeSocketSet(websrv->set);
	websrv->set=NULL;
}

void WebSrv_Destroy(WebSrv *websrv)
{
	WebSrv_Close(websrv);
	free(websrv);
}

#define B64_BLOCK 64

int base64_literal(unsigned char c);
unsigned int MultOf(unsigned int val, unsigned int block)
{
  int mod=val%block;
  int mul=val/block;

  val=block*mul;
  if(mod||(val==0)) val+=block;

  return(val);
}

char *WebSrv_enbase64(void *d, unsigned int datlen)
{
  unsigned char *dat=(unsigned char *)d;
  int strpos=0;
  int strsize=0;
  int pos=0;
  char *str=(char *)malloc(MultOf(datlen,B64_BLOCK));

  if((d==NULL)||(str==NULL))
    return(NULL);

  for(pos=0; pos<datlen; pos++)
  {
    // Resize string if necessary
    if((strpos+3)>=strsize)
    {
      char *str2=(char *)realloc(str,strsize+=B64_BLOCK);
      if(str2==NULL)
      {
        free(str);
        return(NULL);
      }
      str=str2;
    }

    if(base64_literal(dat[pos]))
      str[strpos++]=dat[pos];
    else
      strpos+=sprintf(str+strpos,"=%02x",dat[pos]);
  }

  str[strpos++]='\0';
  return(str);
}

int base64_literal(unsigned char c)
{
  if((c>=33)&&(c<=60))
    return(1);
  else if((c>=62)&&(c<=126))
    return(1);
  else
    return(0);
}


/**
 *  BASE64 decoding routine (C) 1997 Dr. Y. C. Kong.
 *  Full source available at:
 *  http://groups.google.com/groups?selm=5c59fq%24lfs%40oban.cc.ic.ac.uk&oe=UTF-8&output=gplain
 */

/******************************************************/
/*                                                    */
/* Program to decode mime encoding in E-mail messages */
/*                                                    */
/* Supports 1) Hexadecimal encoding (ie =B6=C8=D9=)   */
/*          2) BASE64 encoing (ie ABCDefgh1234Ab0=)   */
/*                                                    */
/* (C) 22/01/97 Dr Y.C. Kong (c.kong@cd.co.uk)        */
/*                                                    */
/* This quick hack allows easy processing of E-mail   */
/* messages with hexadecimal and/or base64. The mail  */
/* header is not read nor is a mime64 identification  */
/* tag needed. The simple algorithm attempts to auto- */
/* matically detect encoded mail. For base64 encoded  */
/* files you need to strip the header and any blank   */
/* lines.                                             */
/*                                                    */
/* The reason that I produced this quick hack and     */
/* worked out the algorithm for base64 is that I am   */
/* feed up with base64 & mime utilities which don't   */
/* work properly. This program is written in 100%     */
/* ANSI C and requires nothing else apart from a C    */
/* compiler. It works under MSDOS, Windows & UNIX     */
/* without a SINGLE change of the source code         */
/* (CR/LF and LF end of line handled internally).     */
/*                                                    */
/* This utility is provided without any warranty.     */
/* The user may use this freely and modify the code   */
/* only if the author name is retained                */
/*                                                    */
/******************************************************/

/*             */
/* Decode line */
/*             */

// Modified to just convert, instead of convert and print.
void base64_decode(char *buffer, char *buffer2, const int maxlen)
{
  int ipnt=0,ipnt2=0;
  int i,num1,num2,num3,num4,temp2,temp3;

/* Try hexadecimal decoding first */
  while (buffer[ipnt] != 0 )
    if (buffer[ipnt] == '=' && ipnt < maxlen-2)
    {
      num2=unhex(buffer[ipnt+1]);
      num1=unhex(buffer[ipnt+2]);
      if (num1 != -1 && num2 != -1)
        buffer[ipnt2++]=num2*16+num1;
      else
        for (i=0; i<3; i++)
          buffer[ipnt2++]=buffer[ipnt+i];
      ipnt+=3;
    }
    else
      buffer[ipnt2++]=buffer[ipnt++];
  buffer[ipnt2]=0;

/* Try base64 decoding if hexadecimal did not work */
  if (ipnt == ipnt2)
  {
    ipnt=0;
    ipnt2=0;
    while (buffer[ipnt] != 0 )
      if (ipnt < maxlen-3)
      {
        num1=unbase64(buffer[ipnt]);
        num2=unbase64(buffer[ipnt+1]);
        num3=unbase64(buffer[ipnt+2]);
        num4=unbase64(buffer[ipnt+3]);
        if (num1 != -1 && num2 != -1 && num3 != -1 && num4 != -1)
/* 010000 010100 001001 000011  QUJD (16,20,9,3) */
/* 01000001 01000010 01000011   ABC (65,66,67)   */
        {
          temp2=num2/16;
          temp3=num3/4;
          buffer2[ipnt2++]=num1*4+temp2;                 /* Byte1 */
          if (buffer[ipnt+2] != '=' )
          {
            buffer2[ipnt2++]=(num2-16*temp2)*16+temp3;   /* Byte2 */
            if (buffer[ipnt+3] != '=' )
              buffer2[ipnt2++]=(num3-4*temp3)*64+num4;   /* Byte3 */
          }
          ipnt+=4;
        }
        else
        {
//          printf("%s\n",buffer);     /* Bad chuck of 4 chars - invalid line */
          return;
        }
      }
      else
      {
//        printf("%s\n",buffer);       /* Hit maxlength - invalid line */
        return;
      }
    buffer2[ipnt2]=0;                /* Line okay - print without newline */
//    if (ipnt != 0)
//      printf("%s",buffer2);
//    else
//      printf("\n");                  /* Empty line */
  }
//  else
//    printf("%s\n",buffer);
}

/*                                */
/* Decode a hexadecimal character */
/*                                */
int unhex(character)
int character;
/*   0 0   1 1   2 2   3 3   4 4   5 5   6 6  7 7   8 8   9 9  */
/*  10 A  11 B  12 C  13 D  14 E  15 F                         */
{
  if (character == '0')
    return(0);
  else
    if (character >= '1' && character <= '9')
      return(1+character-'1');
    else
      if (character >= 'A' && character <= 'F')
        return(10+character-'A');
      else
        return(-1); 
}


/*                         */
/* Decode base64 character */
/*                         */
int unbase64(character)
int character;
/*   0 A   1 B   2 C   3 D   4 E   5 F   6 G   7 H   8 I   9 J  */
/*  10 K  11 L  12 M  13 N  14 O  15 P  16 Q  17 R  18 S  19 T  */
/*  20 U  21 V  22 W  23 X  24 Y  25 Z  26 a  27 b  28 c  29 d  */
/*  30 e  31 f  32 g  33 h  34 i  35 j  36 k  37 l  38 m  39 n  */
/*  40 o  41 p  42 q  43 r  44 s  45 t  46 u  47 v  48 w  49 x  */
/*  50 y  51 z  52 0  53 1  54 2  55 3  56 4  57 5  58 6  59 7  */
/*  60 8  61 9  62 +  63 / pad =                                */
{
  switch (character)
  {
    case '0': return(52);
    case '+': return(62);
    case '/': return(63);
    case '=': return(0);
    default:
      if (character >= 'A' && character <= 'Z')
        return(0+character-'A');
      else
        if (character >= 'a' && character <= 'z')
          return(26+character-'a');
        else
          if (character >= '1' && character <= '9' )
            return(53+character-'1');
          else
            return(-1);
  }
}


// Looks nasty, yeah, but it won't buffer overflow.
char *WebSrv_unbase64(const char *str)
{
  char buf1[256];
  char buf2[256];
  strncpy(buf1,str,256);
  buf1[255]=buf2[255]=buf2[0]='\0';
  base64_decode(buf1,buf2,255);
  return(strdup(buf2));      
}

