#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL_types.h>
#include "ipaddr.h"
#include "main.h"
#include "faf.h"

char *MakeAuth(); // Assenbles random user:pass
enum param_e paramtype(const char *param); // Determines parameter type
static const char  *str_param[]; // Names of the extra specifiers
static unsigned int num_param[]; // Number of parameters they take

void print_usage();

int main(int argc, char *argv[])
{
  int n=1;
  Uint16 port=0;
  char localip[30]="127.0.0.1";
  char *iface=NULL;
  char *fname=NULL;
  char *auth=NULL;

  srand((unsigned int)time(NULL));

  while(n<argc) // Process commandline arguments
  {
    enum param_e p=paramtype(argv[n]);
    switch(p)
    {
    case PARAM_IFACE:
      if(iface!=NULL)
      {
        fprintf(stderr,"Cannot specify more than one interface!\n");
        print_usage();
        return(1);
      }
      else if(argv[n+1]==NULL)
      {
        fprintf(stderr,"No interface specified after %s\n",str_param[PARAM_IFACE]);
        print_usage();
        return(1);
      }

      iface=argv[n+1];
      break;
    case PARAM_AUTH:
      if(auth!=NULL)
      {
        fprintf(stderr,"Can't specify more than one auth!\n");
        print_usage();
        return(1);
      }
      else if(argv[n+1]==NULL)
      {
        fprintf(stderr,"No auth string given after %s!\n",str_param[PARAM_AUTH]);
        print_usage();
        return(1);
      }
      auth=argv[n+1];
      break;

    case PARAM_PORT:
      if(port!=0)
      {
        fprintf(stderr,"Cannot specify more than one port!\n");
        print_usage();
        return(1);
      }
      else if(argv[n+1]==NULL)
      {
        fprintf(stderr,"No port specified after %s!\n",str_param[PARAM_PORT]);
        print_usage();
        return(1);
      }
      port=atoi(argv[n+1]);
      if(port==0)
      {
        fprintf(stderr,"'%s' is not a valid port!\n",argv[n+1]);
        print_usage();
        return(1);
      }
      break;

    case PARAM_FILE:
      if(fname!=NULL)
      {
        fprintf(stderr,"Cannot specify more than one file!\n");
        print_usage();
        return(1);
      }
      else if(access(argv[n],R_OK)!=0)
      {
        fprintf(stderr,"Cannot access file '%s'\n",argv[n]);
        print_usage();
        return(1);
      }
      fname=argv[n];
      break;

    default:
      fprintf(stderr,"Unknown error!\n");
      print_usage();
      return(1);
    }
      // Advance to next commandline parameter
    if(p==PARAM_FILE)  n++;
    else               n+=num_param[p];
  }

  if(fname==NULL)
  {
    fprintf(stderr,"No file specified!\n");
    print_usage();
    return(1);
  }

  if(port==0)
    port=DEFAULT_PORT;
  if(iface==NULL)
    iface=DEFAULT_IFACE;
  if(auth==NULL)
    auth=MakeAuth();

  if(GetIP(iface,localip,NULL)<0)
    fprintf(stderr,"Warning, couldn't get local IP\n");


  if(FAF_Init()<0)
  {
    fprintf(stderr,"Couldn't init FAF\n");
    return(1);
  }

  {
    char *webname=FAF_Webname(fname);
    fprintf(stderr,"Opening http://%s@%s:%d/%s\n",auth,localip,port,webname);
    free(webname);
  }

  if(FAF_Run(auth,fname,port)<0)
  {
    fprintf(stderr,"Error sending file!\n");
    return(1);
  }

  return(0);
}

enum param_e paramtype(const char *param)
{
  enum param_e p;
  for(p=PARAM_AUTH; str_param[p]!=NULL; p++)
    if(strcmp(param,str_param[p])==0)
      return(p);

  return(PARAM_FILE);
}

char *MakeAuth()
{
  static char authbuf[100];
  const char *alphabet="abcdefghijklmnopqrstuvwxyz0123456789";
  int len=strlen(alphabet);
  int n,pos=0;

  for(n=0; n<7; n++)
    authbuf[pos++]=alphabet[rand()%len];

  authbuf[pos++]=':';

  for(n=0; n<7; n++)
    authbuf[pos++]=alphabet[rand()%len];

  authbuf[pos++]='\0';

  return(authbuf);
}

void print_usage()
{
  fprintf(stderr,
    "faf:  File-And-Forget:  Instant temporary webserver, just add water.\n"
    "  The file will be hosted until it's downloaded completely or 2\n"
    "  minutes pass, whichever comes first.  Valid URL will be printed\n"
    "  upon successful launch of server.\n\n"
    "  Usage:  faf [-port 5432] [-auth user:pass] [-iface eth0] filename\n\n");
}

static const char  *str_param[]={"-auth","-port","-iface",NULL};
static unsigned int num_param[]={      2,      2,       2,  -1};

